import datetime
import os
from collections import Counter, OrderedDict
from typing import Dict, Tuple, List

import math

import matplotlib.pyplot as plt
import pandas as pd
import pingouin as pg
import seaborn as sns
from matplotlib.gridspec import GridSpec

from config import VALUE_COLUMN, VIS_FORMAT, PERCEPTION_RATING_COLOR_PALETTE, BACKGROUND_RATING_COLOR_PALETTE, \
    TOP_OFFSET, BOTTOM_OFFSET, SNS_AXES_STYLE, SNS_CONTEXT, FIG_SIZE, RC_FONTS, BOTTOM_BORDER_OFFSET, TOP_BORDER_OFFSET, \
    H_SPACE, DPI, CAPTION_COLOR, BG_COLORS, COMPLEXITY_COLORS
from study1.s1_config import IMAGE_OUTPUT_PATH, VIS_NUMERIC_ORDER, ERROR_WIDTH, VIS_VALUES, \
    VIS_BOOL_ORDER, LLAT, QUESTION_MAPPING, ERROR_COLOR
from study1.base_data_processing import get_questionnaire_data
from study1.data_processing import get_errors, get_times
from utils import add_xaxis_labels, set_legend_from_vis, set_custom_legend, add_yaxis_labels_hue, set_rc_plot_values, \
    add_caption


def background_error_time(error_data: pd.DataFrame, times_data: pd.DataFrame):
    hue_order = ["BG1", "BG2", "BG3"]

    numeric_df = error_data[error_data.valueType == "NUMERIC"]
    bool_df = error_data[error_data.valueType == "BOOL"]
    bool_df[VALUE_COLUMN] = bool_df.apply(lambda row: 1 - row[VALUE_COLUMN], axis=1)

    fig = plt.figure(figsize=(FIG_SIZE[0], FIG_SIZE[1] * 1.1), dpi=DPI)
    gs = GridSpec(1, 15, figure=fig, top=TOP_OFFSET, bottom=BOTTOM_OFFSET / 1.1, left=0.035, right=0.99, wspace=0.9)
    # gs = GridSpec(1, 15, figure=fig)

    sGs = gs.new_subplotspec(loc=[0, 0], rowspan=1, colspan=4).subgridspec(nrows=1, ncols=4, wspace=0.025)
    # Numeric View - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="background", data=numeric_df, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.get_legend().remove()
    ax.get_xaxis().set_visible(False)
    ax.set_xlabel("")
    ax.set_xticklabels(VIS_NUMERIC_ORDER)
    ax.set_ylabel("Absolute Errors")
    ax.set_ylim(VIS_VALUES["EN-ylim"])
    ax.set_yticks(VIS_VALUES["EN-yTicks"])
    ax.set_yticklabels(VIS_VALUES["EN-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Numeric View - Analysis Tasks
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="background", data=numeric_df, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=VIS_NUMERIC_ORDER, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel(None)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VIS_VALUES["EN-ylim"])
    ax.set_yticks(VIS_VALUES["EN-yTicks"])
    ax.set_yticklabels([None] * len(VIS_VALUES["EN-yLabels"]))
    add_xaxis_labels(ax, labels=VIS_NUMERIC_ORDER, pos="bottom")

    sGs = gs.new_subplotspec(loc=[0, 4], rowspan=1, colspan=4).subgridspec(nrows=1, ncols=4, wspace=0.025)
    # Bool View - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="background", data=bool_df, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Percentage Error")
    ax.set_xticklabels(VIS_BOOL_ORDER)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VIS_VALUES["EB-ylim"])
    ax.set_yticks(VIS_VALUES["EB-yTicks"])
    ax.set_yticklabels(VIS_VALUES["EB-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Bool View - Analysis Tasks
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="background", data=bool_df, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=VIS_BOOL_ORDER, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VIS_VALUES["EB-ylim"])
    ax.set_yticks(VIS_VALUES["EB-yTicks"])
    ax.set_yticklabels([""] * len(VIS_VALUES["EB-yLabels"]))
    add_xaxis_labels(ax, labels=VIS_BOOL_ORDER, pos="bottom")

    times_data = times_data[times_data.note == "SolvingDuration"]

    sGs = gs.new_subplotspec(loc=[0, 8], rowspan=1, colspan=8).subgridspec(nrows=1, ncols=8, wspace=0.025)
    # Time View - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="background", data=times_data, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Completion Time in s")
    ax.set_xticklabels(LLAT)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VIS_VALUES["T-ylim"])
    ax.set_yticks(VIS_VALUES["T-yTicks"])
    ax.set_yticklabels(VIS_VALUES["T-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Time View - Analysis Tasks
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="questionType", hue="background", data=times_data, ax=ax,
                    palette={k: BG_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=LLAT, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.set_xticklabels(LLAT)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VIS_VALUES["T-ylim"])
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_yticks(VIS_VALUES["T-yTicks"])
    ax.set_yticklabels([""] * len(VIS_VALUES["T-yLabels"]))
    add_xaxis_labels(ax, labels=LLAT, pos="bottom")

    add_caption(fig, "A")
    set_legend_from_vis(fig, ax)
    fig.tight_layout()

    for format_str in VIS_FORMAT:
        fig.savefig(f"{IMAGE_OUTPUT_PATH}/background-questionType.{format_str}")
    plt.close(fig)


def background_complexity(error_data: pd.DataFrame, times_data: pd.DataFrame):
    hue_order = list(COMPLEXITY_COLORS.keys())
    background_order = ["BG1", "BG2", "BG3"]

    numeric_df = error_data[error_data.valueType == "NUMERIC"]
    bool_df = error_data[error_data.valueType == "BOOL"]
    bool_df[VALUE_COLUMN] = bool_df.apply(lambda row: 1 - row[VALUE_COLUMN], axis=1)

    fig = plt.figure(figsize=(FIG_SIZE[0], FIG_SIZE[1] * 1.1), dpi=DPI)
    # fig = plt.figure(figsize=(FIG_SIZE[0] * 12. / 17., FIG_SIZE[1]), dpi=DPI)
    gs = GridSpec(1, 12, figure=fig, top=TOP_OFFSET, bottom=BOTTOM_OFFSET / 1.1, left=0.02, right=0.99, wspace=0.65)

    sGs = gs.new_subplotspec(loc=[0, 0], rowspan=1, colspan=4).subgridspec(nrows=1, ncols=4, wspace=0.025)
    # Numeric View - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="complexity", data=numeric_df, ax=ax,
                    palette={k: COMPLEXITY_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Absolute Errors")
    ax.set_xticklabels(background_order)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VIS_VALUES["EN-C-ylim"])
    ax.set_yticks(VIS_VALUES["EN-C-yTicks"])
    ax.set_yticklabels(VIS_VALUES["EN-C-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Add significant asterisks to the vis
    for i, (source, target) in enumerate([(1, 4), (1, 3), (2, 4), (3, 5), (2, 3), (4, 5)]):
    # for i, (source, target) in enumerate([(1, 4), (3, 5), (1, 3), (4, 5), (2, 4), (2, 3)]):
        b1 = ax.containers[source - 1][0]
        b2 = ax.containers[target - 1][0]
        height = ax.get_ylim()[1]
        x1 = b1.get_x() + b1.get_width() / 2.
        x2 = b2.get_x() + b2.get_width() / 2.
        y = b1.get_y() + height * 0.015 + height * 0.0522 * i
        ax.plot([x1, x2], [y, y], linewidth=2, c="black")
        ax.text(x1 + (x2 - x1) / 2., y + 0.0725, "*", ha="center", va="center", color="black", weight="bold")

    # Numeric View - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="background", hue="complexity", data=numeric_df, ax=ax,
                    palette={k: COMPLEXITY_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=background_order, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(background_order)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VIS_VALUES["EN-C-ylim"])
    ax.set_yticks(VIS_VALUES["EN-C-yTicks"])
    ax.set_yticklabels([""] * len(VIS_VALUES["EN-C-yLabels"]))
    add_xaxis_labels(ax, labels=background_order, pos="bottom")

    sGs = gs.new_subplotspec(loc=[0, 4], rowspan=1, colspan=4).subgridspec(nrows=1, ncols=4, wspace=0.025)
    # Bool View - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="complexity", data=bool_df, ax=ax,
                    palette={k: COMPLEXITY_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Percentage Error")
    ax.set_xticklabels(background_order)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VIS_VALUES["EB-ylim"])
    ax.set_yticks(VIS_VALUES["EB-yTicks"])
    ax.set_yticklabels(VIS_VALUES["EB-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Add significant asterisks to the vis
    for i, (source, target, text) in enumerate([(1, 5, "**"), (2, 5, "***"), (2, 4, "*"), (3, 5, "*"), (3, 4, "*")]):
        b1 = ax.containers[source - 1][0]
        b2 = ax.containers[target - 1][0]
        height = ax.get_ylim()[1]
        x1 = b1.get_x() + b1.get_width() / 2.
        x2 = b2.get_x() + b2.get_width() / 2.
        y = b1.get_y() + height * 0.014 + height * 0.069 * i
        ax.plot([x1, x2], [y, y], linewidth=2, c="black")
        ax.text(x1 + (x2 - x1) / 2., y + 0.0075, text, ha="center", va="center", color="black", weight="bold")

    # Bool View - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="background", hue="complexity", data=bool_df, ax=ax,
                    palette={k: COMPLEXITY_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=background_order, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(background_order)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VIS_VALUES["EB-ylim"])
    ax.set_yticks(VIS_VALUES["EB-yTicks"])
    ax.set_yticklabels([""] * len(VIS_VALUES["EB-yLabels"]))
    add_xaxis_labels(ax, labels=background_order, pos="bottom")

    times_data = times_data[times_data.note == "SolvingDuration"]

    sGs = gs.new_subplotspec(loc=[0, 8], rowspan=1, colspan=4).subgridspec(nrows=1, ncols=4, wspace=0.025)
    # Time - Total
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 0])
        sns.barplot(y=VALUE_COLUMN, x="valueType", hue="complexity", data=times_data, ax=ax,
                    palette={k: COMPLEXITY_COLORS[k] for k in hue_order},
                    hue_order=hue_order, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("Completion Time in s")
    ax.set_xticklabels(background_order)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VIS_VALUES["T-ylim"])
    ax.set_yticks(VIS_VALUES["T-yTicks"])
    ax.set_yticklabels(VIS_VALUES["T-yLabels"])
    add_xaxis_labels(ax, labels=["Total"], pos="bottom")

    # Time - Analysis Task
    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(sGs[0, 1:])
        sns.barplot(y=VALUE_COLUMN, x="background", hue="complexity", data=times_data, ax=ax,
                    palette={k: COMPLEXITY_COLORS[k] for k in hue_order},
                    hue_order=hue_order, order=background_order, capsize=0.075, errwidth=ERROR_WIDTH, errcolor=ERROR_COLOR)
    ax.get_legend().remove()
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticklabels(background_order)
    ax.get_xaxis().set_visible(False)
    ax.set_ylim(VIS_VALUES["T-ylim"])
    ax.set_yticks(VIS_VALUES["T-yTicks"])
    ax.set_yticklabels([""] * len(VIS_VALUES["T-yLabels"]))
    add_xaxis_labels(ax, labels=background_order, pos="bottom")
    # add_label_to_bars(times_data, ax, background_order, hue_key="complexity", x_order_key="background", value_key=VALUE_COLUMN, rotation=90)

    add_caption(fig, "B")
    set_legend_from_vis(fig, ax)
    fig.tight_layout()
    for format_str in VIS_FORMAT:
        fig.savefig(f"{IMAGE_OUTPUT_PATH}/background-complexity.{format_str}")
    plt.close(fig)


def background_user_ratings(questionnaire_df: pd.DataFrame):
    questionnaire_df = questionnaire_df[[f"backgRating{i}" for i in range(1, 5)]]
    questionnaire_df = questionnaire_df.replace("B", "FC=2.53").replace("C", "FC=2.63").replace("D", "FC=3.36").replace("A", "FC=1.65")

    df_values = {"FC=2.63": [], "FC=3.36": [], "FC=2.53": [], "FC=1.65": []}
    for _, row in questionnaire_df.iterrows():
        for i, value in enumerate(row.values):
            df_values[value].append(i + 1)

    values_df = pd.DataFrame(df_values)
    rating_counts = {"background": [], 1: [], 2: [], 3: [], 4: []}
    for b in df_values.keys():
        rating_counts["background"].append(b)
        rating_counts[1].append(values_df[b][values_df[b].isin(range(1, 2))].count())
        rating_counts[2].append(values_df[b][values_df[b].isin(range(1, 3))].count())
        rating_counts[3].append(values_df[b][values_df[b].isin(range(1, 4))].count())
        rating_counts[4].append(values_df[b][values_df[b].isin(range(1, 5))].count())

    rating_count_df = pd.DataFrame(rating_counts, columns=rating_counts.keys())

    fig = plt.figure(figsize=(FIG_SIZE[0] / 2., FIG_SIZE[1]), dpi=DPI)
    gs = GridSpec(1, 1, figure=fig, top=TOP_OFFSET, bottom=BOTTOM_OFFSET - 0.025 / (2 * 5 / 7.), left=0.085, right=0.985)

    with sns.axes_style(SNS_AXES_STYLE):
        ax = fig.add_subplot(gs[0, 0])

        for i in range(4, 0, -1):
                sns.barplot(y="background", x=i, data=rating_count_df,
                            ax=ax, palette=[BACKGROUND_RATING_COLOR_PALETTE[i - 1]])

    # ax.set_xlabel("Number of Votes")
    ax.set_ylabel("")
    ax.set_xlabel("")
    ax.set_xlim([0, 18])
    # ax.set_xticks(range(0, 19, 3))
    ax.set_xticks(range(0, 19))
    ax.set_xticklabels([i if i % 3 == 0 else "" for i in range(0, 19)])

    set_custom_legend(
        fig=fig,
        ax=ax,
        keys=["1", "2", "3", "4"],
        colors=BACKGROUND_RATING_COLOR_PALETTE,
        title="Rating"
    )

    # add_caption(fig, "C")
    fig.tight_layout()
    fig.subplots_adjust(top=TOP_OFFSET, bottom=BOTTOM_OFFSET)
    for format_str in VIS_FORMAT:
        fig.savefig(f"{IMAGE_OUTPUT_PATH}/stoerfaktor.{format_str}")
    plt.close(fig)


def questionnaire_general_perception(questionnaire_df: pd.DataFrame):
    questions = [f"generalQuestion{i}" for i in range(1, 4)]

    fig = plt.figure(figsize=(FIG_SIZE[0] / 2., FIG_SIZE[1] * (2.5 * 5 / 7.)), dpi=DPI)
    gs = GridSpec(4, 1, figure=fig, top=(1 - (1 - TOP_OFFSET) / (2 * 5 / 7.)), right=0.99, left=0.035, bottom=(BOTTOM_OFFSET - 0.025) / (2 * 5 / 7.),
                  hspace=H_SPACE * 0.8)

    question_counts = {"background": [], "question": [], 1: [], 2: [], 3: [], 4: [], 5: []}
    for b in questionnaire_df.background.unique():
        for qi, question in enumerate(questions):
            counts = questionnaire_df[questionnaire_df.background == b][question].value_counts()
            counts = [counts[i] if i in counts else 0 for i in range(1, 6)]

            question_counts["background"].append(b)
            question_counts["question"].append(question)
            question_counts[1].append(int(sum(counts[4:5])))
            question_counts[2].append(int(sum(counts[3:5])))
            question_counts[3].append(int(sum(counts[2:5])))
            question_counts[4].append(int(sum(counts[1:5])))
            question_counts[5].append(int(sum(counts[0:5])))

    question_count_df = pd.DataFrame(question_counts, columns=question_counts.keys())

    axes = []
    with sns.axes_style(SNS_AXES_STYLE):
        for i, question in enumerate(questions):
            ax = fig.add_subplot(gs[i, 0])

            for j in range(5, 0, -1):
                sns.barplot(y="question", x=j, hue="background",
                            data=question_count_df[question_count_df.question == question],
                            ax=ax, hue_order=["BG1", "BG2", "BG3"],
                            palette=[PERCEPTION_RATING_COLOR_PALETTE[j - 1] for _ in range(0, 3)])

            ax.set_xlim((0, 6.25))
            ax.set_xticks(range(0, 7))
            ax.set_xticklabels(["" for _ in range(0, 7)])
            ax.set_yticklabels([""])
            ax.get_legend().remove()
            ax.set_ylabel("")
            ax.set_xlabel("")
            ax.set_title(QUESTION_MAPPING[question], loc="left")

            add_yaxis_labels_hue(ax, labels=["BG1", "BG2", "BG3"], y_offset=-0.05, fontsize=24)
            axes.append(ax)

        ax.set_xticklabels(range(0, 7))

    for i, ax in enumerate([axes[c] for c in [1, 2]]):
        b0 = ax.containers[0][0]
        b1 = ax.containers[1][0]
        b2 = ax.containers[2][0]
        x0 = b0.get_width()
        x1 = x0 + 0.03
        x2 = x0 + 0.14

        y0 = b0.get_y() + b0.get_height() / 2
        y1 = b1.get_y() + b0.get_height() / 2
        y20 = b2.get_y() + b2.get_height() / 2 + 0.04
        y21 = b2.get_y() + b2.get_height() / 2 - 0.04

        ax.plot([x0, x1, x1, x0], [y1, y1, y21, y21], linewidth=2, c="black", scaley=False, scalex=False)
        ax.text(x1 + 0.015, y1 / 2., "**" if i is 3 else "*", ha='left', va='top', color="black", weight="bold")
        ax.plot([x0, x2, x2, x0], [y0, y0, y20, y20], linewidth=2, c="black", scaley=False, scalex=False)
        ax.text(x2 + 0.015, y0, "**" if i is 3 else "*", ha='left', va='top', color="black", weight="bold")

    add_caption(fig, "A")
    set_custom_legend(
        fig=fig,
        ax=ax,
        keys=["agree", "", "neutral", "", "disagree"],
        colors=PERCEPTION_RATING_COLOR_PALETTE,
        title="Five-Point Scale Answers"
    )

    fig.tight_layout()
    fig.subplots_adjust(top=TOP_OFFSET)
    for format_str in VIS_FORMAT:
        fig.savefig(f"{IMAGE_OUTPUT_PATH}/questionnaire-general-perception-rating.{format_str}")
    plt.close(fig)


def questionnaire_vis_perception(questionnaire_df: pd.DataFrame):
    questions = [f"{k}{i}" for k in ["generalPerception", "backgPerception"] for i in range(1, 3)]

    fig = plt.figure(figsize=(FIG_SIZE[0] / 2., FIG_SIZE[1] * (2.5 * 5 / 7.)), dpi=DPI)
    gs = GridSpec(4, 1, figure=fig, top=(1 - (1 - TOP_OFFSET) / (2 * 5 / 7.)), right=0.99, left=0.035, bottom=(BOTTOM_OFFSET - 0.025) / (2 * 5 / 7.),
                  hspace=H_SPACE * 0.8)

    question_counts = {"background": [], "question": [], "mean": [], 1: [], 2: [], 3: [], 4: [], 5: []}
    for b in questionnaire_df.background.unique():
        for qi, question in enumerate(questions):
            counts = questionnaire_df[questionnaire_df.background == b][question].value_counts()
            mean = questionnaire_df[questionnaire_df.background == b][question].mean()
            counts = [counts[i] if i in counts else 0 for i in range(1, 6)]

            question_counts["background"].append(b)
            question_counts["question"].append(question)
            question_counts["mean"].append(mean)
            question_counts[1].append(int(sum(counts[4:5])))
            question_counts[2].append(int(sum(counts[3:5])))
            question_counts[3].append(int(sum(counts[2:5])))
            question_counts[4].append(int(sum(counts[1:5])))
            question_counts[5].append(int(sum(counts[0:5])))

    question_count_df = pd.DataFrame(question_counts, columns=question_counts.keys())

    axes = []
    with sns.axes_style(SNS_AXES_STYLE):
        for i, question in enumerate(questions):
            ax = fig.add_subplot(gs[i, 0])

            for j in range(5, 0, -1):
                sns.barplot(y="question", x=j, hue="background",
                            data=question_count_df[question_count_df.question == question],
                            ax=ax, hue_order=["BG1", "BG2", "BG3"],
                            palette=[PERCEPTION_RATING_COLOR_PALETTE[j - 1] for _ in range(0, 3)])

            ax.set_xlim((0, 6.25))
            ax.set_xticks(range(0, 7))
            ax.set_xticklabels(["" for _ in range(0, 7)])
            ax.set_yticklabels([""])
            ax.get_legend().remove()
            ax.set_ylabel("")
            ax.set_xlabel("")
            ax.set_title(QUESTION_MAPPING[question], loc="left")

            add_yaxis_labels_hue(ax, y_offset=-0.05, fontsize=24,
                                 labels=[
                                     f'BG1',
                                     f'BG2',
                                     f'BG3'
                                 ])
            axes.append(ax)

        ax.set_xticklabels(range(0, 7))

    # Add significance lines and asterisks
    b0 = axes[2].containers[0][0]
    b2 = axes[2].containers[2][0]
    x0 = b0.get_width()
    x2 = x0 + 0.03
    y0 = b0.get_y() + b0.get_height() / 2
    y2 = b2.get_y() + b2.get_height() / 2

    axes[2].plot([x0, x2, x2, x0], [y0, y0, y2, y2], linewidth=2, c="black", scaley=False, scalex=False)
    axes[2].text(x2 + 0.015, y0, "*", ha='left', va='top', color="black", weight="bold")

    for i, ax in enumerate([axes[c] for c in [1, 3]]):
        b0 = ax.containers[0][0]
        b1 = ax.containers[1][0]
        b2 = ax.containers[2][0]
        x0 = b0.get_width()
        x1 = x0 + 0.03
        x2 = x0 + 0.14

        y0 = b0.get_y() + b0.get_height() / 2
        y1 = b1.get_y() + b0.get_height() / 2
        y20 = b2.get_y() + b2.get_height() / 2 + 0.04
        y21 = b2.get_y() + b2.get_height() / 2 - 0.04

        ax.plot([x0, x1, x1, x0], [y1, y1, y21, y21], linewidth=2, c="black", scaley=False, scalex=False)
        ax.text(x1 + 0.015, y1 / 2., "**" if i is 3 else "*", ha='left', va='top', color="black", weight="bold")
        ax.plot([x0, x2, x2, x0], [y0, y0, y20, y20], linewidth=2, c="black", scaley=False, scalex=False)
        ax.text(x2 + 0.015, y0, "**" if i is 3 else "*", ha='left', va='top', color="black", weight="bold")

    set_custom_legend(
        fig=fig,
        ax=ax,
        keys=["greatly", "", "moderate", "", "not at all"],
        colors=PERCEPTION_RATING_COLOR_PALETTE,
        title="Five-Point Scale Answers"
    )

    add_caption(fig, "B")
    fig.tight_layout()
    fig.subplots_adjust(top=TOP_OFFSET)
    for format_str in VIS_FORMAT:
        fig.savefig(f"{IMAGE_OUTPUT_PATH}/questionnaire-vis-perception-rating.{format_str}")
    plt.close(fig)


if __name__ == '__main__':
    os.makedirs(IMAGE_OUTPUT_PATH, exist_ok=True)

    set_rc_plot_values()

    filter_methods = [
        lambda df: df[df.blockType != "T"],
        lambda df: df.replace("L", "BG1").replace("S+F", "BG2").replace("A", "BG3"),
        lambda df: df.replace("Extremum", LLAT[3]).replace("Range Values", LLAT[4]).replace("Compute Derived Value", LLAT[2]).replace("Find Anomalies", LLAT[5]).replace("Retrieve Value", LLAT[0])
    ]

    errors_df = get_errors(
        filter_methods=[lambda df: df[df.questionType != LLAT[4]]] + filter_methods + [
            lambda df: df[~df.questionType.isin(["Extreme Values", "Delta", "Min", "Max"])]
        ],
        reload_csv=False
    )
    errors_df["complexity"] = errors_df.apply(lambda row: list(COMPLEXITY_COLORS.keys())[row["complexity"] - 1], axis=1)
    times_df = get_times(
        filter_methods=filter_methods + [
            lambda df: df.replace("Determine Range", LLAT[4])
        ],
        reload_csv=False
    )
    times_df["complexity"] = times_df.apply(lambda row: list(COMPLEXITY_COLORS.keys())[row["complexity"] - 1], axis=1)

    questionnaire_df = get_questionnaire_data()

    background_error_time(errors_df, times_df)
    background_complexity(errors_df, times_df)
    background_user_ratings(questionnaire_df)

    questionnaire_general_perception(questionnaire_df)
    questionnaire_vis_perception(questionnaire_df)

